"""Top-level functions that provide UI views and abstract away the underlying machinery.

This module is star-imported into `__init__.py` (i.e. `from uikitty.functions import *`)
and therefore effectively serves as the public API surface for this package.

All functions in this module must be thoroughly documented and (relatively) easy to use,
and should (if feasible/applicable) provide reasonable options for user customization.
"""

from collections.abc import Callable
from typing import Any

from discord import ApplicationContext, ButtonStyle, Embed

from uikitty.dynamic_selector import DynamicSelector


async def dynamic_select(
    ctx: ApplicationContext,
    *args: str,
    content: str | None = None,
    embed: Embed | None = None,
    button_style: ButtonStyle = ButtonStyle.secondary,
    select_placeholder: str | None = None,
    log: Callable[[str], None] | None = print,
    **kwargs: Any,
) -> Any:
    """Displays a collection of options and returns the one selected by the user.

    Note:
        The appearance/behavior of the view rendered by this function (if any)
        is determined by the number of options that are provided:

        - If there are **no** options, this function will simply return `None`.

        - If there is **exactly 1** option, this function will return that option (if
          specified via `*args`) or the option's value (if specified via `**kwargs`).

        - If there are **between 2 and 5** options (inclusive), they will be displayed
          as a single row of `Button` components (one button per option). This function
          will return the selected option (or its value) once the user clicks a button.

        - If there are **between 6 and 25** options (inclusive), they will be displayed
          in a single `Select` component. This function will return the selected option
          (or its value) once the user clicks an item in the dropdown menu.

        - If there are **26 or more** options, they will be evenly divided into pages
          and displayed in a composite `Paginator` view. This function will return the
          selected option (or its value) once the user selects it (using the `Select`
          component) and confirms their choice (using the central `Button` component).

    Args:
        ctx:
            The context for the application command that prompted this selection.
        *args:
            An ordered collection of option strings.
            Cannot be used alongside `**kwargs`.
        content:
            The text content to display in the message with the selector view.
            Optional. May be used alongside `embed`.
        embed:
            An embed to display in the message with the selector view.
            Optional. May be used alongside `content`.
        button_style:
            The `ButtonStyle` to use for the option `Button` components.
            Only applicable if there are between 2 and 5 options (inclusive).
        select_placeholder:
            The placeholder text to display in the `Select` (dropdown) component.
            Only applicable if there are between 6 and 25 options (inclusive).
        log:
            A function that will be called to display debug information in the console.
            Set to `None` to disable console output for this function call.
        **kwargs:
            An ordered mapping of option labels to option values.
            Cannot be used alongside `*args`.

    Returns:
        The user's selected option string (if `*args`) or option value (if `**kwargs`).
    """
    if args and kwargs:
        raise ValueError(
            "Either '*args' or '**kwargs' (not both!) "
            "must be provided to define the available options."
        )

    options = kwargs if kwargs else {arg: arg for arg in args}

    if len(options) < DynamicSelector.MIN_OPTIONS:
        return next(iter(options.values())) if options else None

    view = DynamicSelector(ctx, button_style, select_placeholder, log, options)
    respond = ctx.edit if ctx.response.is_done() else ctx.respond

    await respond(content=content, embed=embed, view=view)
    await view.wait()

    return view.selected_value



"""This script simplifies the project README to avoid confusing PyPI with formatting."""

import re
from pathlib import Path


def get_github_link(header: str) -> str:
    return (
        f"https://github.com/nuztalgia/bot-ui-kitty#{header.lower().replace(' ', '-')}"
    )


def main() -> int:
    readme_file = (Path(__file__) / "../../README.md").resolve()
    readme_text = readme_file.read_text()

    # Currently, "Available Views" is the only section that needs to be simplified.
    available_views_match = re.search(
        r"## Available Views\n\n(.+)\n## [A-Z]", readme_text, flags=re.DOTALL
    )

    if not available_views_match:
        print("Could not find 'Available Views' section in 'README.md'.")
        return 1

    original_text = (txt := available_views_match.group(1))[: txt.rindex("\n ")].strip()

    # The header for each view type will link to the corresponding header on GitHub.
    partially_simplified_text = re.sub(
        r"^### ([a-zA-Z ]+)$",
        lambda match: f"### [{match.group(1)}]({get_github_link(match.group(1))})",
        original_text,
        flags=re.MULTILINE,
    )

    # Each "Example" bullet point will only contain a single image and a code snippet.
    fully_simplified_text = re.sub(
        r"(- \*+Example.+?)\n.+?(https://user-images.+?\.png).+?```",
        lambda match: f"{match.group(1)}\n\n  ![image]({match.group(2)})\n\n  ```",
        partially_simplified_text,
        flags=re.DOTALL,
    )

    readme_file.write_text(readme_text.replace(original_text, fully_simplified_text))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import json
from typing import Any, Final

from discord import ApplicationContext, Bot, Cog, Embed, SlashCommandGroup

import uikitty


class DynamicSelectCog(Cog):
    def __init__(self, bot: Bot) -> None:
        self.bot: Final[Bot] = bot
        self.elements_data: Final[dict[str, Any]] = {}

    select = SlashCommandGroup("select", "Examples for the Dynamic Select view.")

    @select.command(
        description="An example of a Dynamic Select view with 3 options.",
    )
    async def pokemon(self, ctx: ApplicationContext) -> None:
        number = await uikitty.dynamic_select(
            ctx,
            content="Choose your starter Pokemon!",
            **{  # type: ignore[arg-type]
                "🌱 Bulbasaur": 1,
                "🔥 Charmander": 4,
                "💧 Squirtle": 7,
            },
        )
        await ctx.edit(content=f"Your Pokemon's number is **#00{number}**!", view=None)

    @select.command(
        description="An example of a Dynamic Select view with 24 options.",
    )
    async def time(self, ctx: ApplicationContext) -> None:
        time = await uikitty.dynamic_select(
            ctx,
            *[f"{str(i).zfill(2)}:00" for i in range(24)],
            select_placeholder="What time is it, Mr. Wolf?",
        )
        await ctx.edit(content=f"It's ~~{time}~~ **DINNER TIME!!!**", view=None)

    @select.command(description="An example of a Dynamic Select view with 119 options.")
    async def element(self, ctx: ApplicationContext) -> None:
        if not self.elements_data:
            await ctx.response.defer()
            await self._fetch_elements()

        color = ctx.guild.me.color
        element = await uikitty.dynamic_select(
            ctx,
            embed=Embed(title="Select an element to learn more about it!", color=color),
            **self.elements_data,
        )
        embed = Embed(
            title=element["name"],
            description=element["summary"],
            url=element["source"],
            color=color,
        )
        await ctx.edit(embed=embed, view=None)

    async def _fetch_elements(self) -> None:
        data_url = (
            "https://raw.githubusercontent.com/Bowserinator/"
            "Periodic-Table-JSON/master/PeriodicTableJSON.json"
        )
        # noinspection PyUnresolvedReferences, PyProtectedMember
        async with self.bot.http._HTTPClient__session.get(data_url) as response:
            elements = json.loads(await response.text())["elements"]

        for element in elements:
            label = f"{element['number']}. {element['name']} ({element['symbol']})"
            self.elements_data[label] = {
                k: v for k, v in element.items() if k in {"name", "source", "summary"}
            }


def setup(bot: Bot) -> None:
    bot.add_cog(DynamicSelectCog(bot))



from collections.abc import Callable
from typing import Any, Final

from discord import Interaction
from discord.ui import View


class BaseSelector(View):
    """Base class for views that let the user pick from a set of predefined options."""

    def __init__(
        self,
        log: Callable[[str], None] | None,
        options: dict[str, Any],
    ) -> None:
        """Initializes a new `BaseSelector` instance.

        Note:
            In practice, this should only be called via `super()` from subclasses.

        Args:
            log:
                A function that will be called to display information in the console.
                Set to `None` to disable console output for this instance.
            options:
                A dictionary that maps option keys (i.e. names/labels) to option values.
        """
        super().__init__(timeout=None)

        def _log(message: str) -> None:
            if log:
                log(message)

        self.log: Final[Callable[[str], None]] = _log
        self.options: Final[dict[str, Any]] = options
        self.selected_value: Any = None

    async def finish(self, selected_key: str, interaction: Interaction) -> None:
        """Sets this instance's `selected_value` to the value for the `selected_key`.

        Subclasses should rely on this method to call `stop()` and ensure
        that an appropriate value is available for subsequent consumers.

        Args:
            selected_key:
                The `options` dictionary key for the item that was selected by the user.
            interaction:
                The `Interaction` on the UI component that triggered this method call.

        Raises:
            KeyError: If `selected_key` is not in this instance's `options` dictionary.
        """
        await interaction.response.defer()
        self.log(f"└── {interaction.user} selected '{selected_key}'.")
        self.selected_value = self.options[selected_key]
        self.stop()



import functools
from collections.abc import Callable, Coroutine
from typing import Any, Final, NamedTuple

from discord import ApplicationContext, ButtonStyle, Interaction, SelectOption
from discord.ui import Button, Select

from uikitty.base_selector import BaseSelector


class Paginator:
    """A helper class that displays and manages paginated options for a `BaseSelector`.

    This class must be initialized with multiple lists of `SelectOption` items, in which
    each `list` represents the options that should be rendered on an individual page.

    After an instance of this class is created, its `attach()` method must be called in
    order to make it functional. The provided `BaseSelector`'s `finish()` method will be
    called asynchronously once the user selects an option and confirms their choice.
    """

    class UI(NamedTuple):
        """A `NamedTuple` containing the individual UI components in a `Paginator`."""

        select: Select = Select()
        """A dropdown menu containing all the available options on the current page."""

        prev_button: Button = Button(style=ButtonStyle.primary, label="<<")
        """Navigates to the previous page when clicked. Disabled when on first page."""

        center_button: Button = Button()
        """Confirms the selected option when clicked. Disabled if no option is selected.
           While disabled, will show information about the current pagination state."""

        next_button: Button = Button(style=ButtonStyle.primary, label=">>")
        """Navigates to the next page when clicked. Disabled when on last page."""

    def __init__(
        self,
        *options_by_page: list[SelectOption],
        placeholder: str = "Make a selection, or use the arrows for more options",
    ) -> None:
        """Initializes a new `Paginator` instance.

        Args:
            *options_by_page:
                Lists of `SelectOption` items, in which each `list` represents
                the options that should be rendered on an individual page.
            placeholder:
                The placeholder text to display in the `Select` (dropdown) menu.

        Raises:
            ValueError: If fewer than `2` lists (i.e. pages) of options are provided.
        """
        if len(options_by_page) < 2:
            raise ValueError("At least two lists (i.e. pages) of options are required.")

        self.options_by_page: Final[tuple[list[SelectOption], ...]] = options_by_page
        self.default_placeholder: Final[str] = placeholder
        self.ui: Final[Paginator.UI] = type(self).UI()

        self.ui.select.callback = self._on_select
        self.ui.prev_button.callback = self._on_prev_click
        self.ui.next_button.callback = self._on_next_click

        async def uninitialized() -> None:
            raise RuntimeError("'self.refresh_view' was never properly initialized.")

        self.refresh_view: Callable[[], Coroutine[Any, Any, None]] = uninitialized
        self.current_page: int = 0
        self.current_selection: str | None = None

    def attach(self, parent_view: BaseSelector, ctx: ApplicationContext) -> None:
        """Adds pagination UI components to the view and sets up the required callbacks.

        Args:
            parent_view:
                The `BaseSelector` view to which these UI components will be added.
                Once a selection is made, this view's `finish()` method will be called.
            ctx:
                The context for the application command that prompted this selection.
        """

        async def on_confirm(interaction: Interaction) -> None:
            if isinstance(self.current_selection, str):
                await parent_view.finish(self.current_selection, interaction)
            else:
                raise TypeError("Cannot confirm a nonexistent selection.")

        self.refresh_view = functools.partial(ctx.edit, view=parent_view)
        self.ui.center_button.callback = on_confirm
        self._update_ui()

        for component in self.ui:
            parent_view.add_item(component)

    @property
    def _has_selection(self) -> bool:
        return self.current_selection is not None

    async def _on_select(self, interaction: Interaction) -> None:
        self.current_selection = self.ui.select.values[0]
        await self._on_interaction(interaction)

    async def _on_prev_click(self, interaction: Interaction) -> None:
        self.current_page -= 1
        self.current_selection = None
        await self._on_interaction(interaction)

    async def _on_next_click(self, interaction: Interaction) -> None:
        self.current_page += 1
        self.current_selection = None
        await self._on_interaction(interaction)

    async def _on_interaction(self, interaction: Interaction) -> None:
        await interaction.response.defer()
        self._update_ui()
        await self.refresh_view()

    def _update_ui(self) -> None:
        page_count = len(self.options_by_page)
        options = self.options_by_page[self.current_page]
        center_button_label = f"Page {self.current_page + 1} of {page_count}"

        if self._has_selection:
            self.ui.select.placeholder = self.current_selection
            self.ui.select.options = [
                option for option in options if (option.label != self.current_selection)
            ]
            self.ui.center_button.style = ButtonStyle.success
            self.ui.center_button.label = "Confirm Selection"
            self.ui.center_button.disabled = False
        else:
            self.ui.select.placeholder = self.default_placeholder
            self.ui.select.options = options
            self.ui.center_button.style = ButtonStyle.secondary
            self.ui.center_button.label = center_button_label
            self.ui.center_button.disabled = True

        self.ui.prev_button.disabled = self.current_page == 0
        self.ui.next_button.disabled = self.current_page == (page_count - 1)



import functools
import math
from collections.abc import Callable
from typing import Any, Final

from discord import ApplicationContext, ButtonStyle, Interaction, SelectOption
from discord.ui import Button, Select

from uikitty.base_selector import BaseSelector
from uikitty.paginator import Paginator


class DynamicSelector(BaseSelector):
    """A `BaseSelector` view whose appearance is determined by its number of options."""

    MAX_ENTRIES_PER_PAGE: Final[int] = 25
    MAX_BUTTONS_PER_ROW: Final[int] = 5
    MIN_OPTIONS: Final[int] = 2

    def __init__(
        self,
        ctx: ApplicationContext,
        button_style: ButtonStyle,
        select_placeholder: str | None,
        log: Callable[[str], None] | None,
        options: dict[str, Any],
    ) -> None:
        """Initializes a new `DynamicSelector` instance.

        Args:
            ctx:
                The context for the application command that prompted this selection.
            button_style:
                The `ButtonStyle` to use for the option `Button` components.
                Only applicable if there are between 2 and 5 options (inclusive).
            select_placeholder:
                The placeholder text to display in the `Select` (dropdown) component.
                Only applicable if there are between 6 and 25 options (inclusive).
            log:
                A function that will be called to display information in the console.
                Set to `None` to disable console output for this instance.
            options:
                A dictionary that maps option keys (i.e. names/labels) to option values.

        Raises:
            ValueError: If fewer than `2` options are provided.
        """
        super().__init__(log=log, options=options)
        cls = type(self)

        match len(self.options):
            case n if n > cls.MAX_ENTRIES_PER_PAGE:
                page_count = math.ceil(n / cls.MAX_ENTRIES_PER_PAGE)
                self.log(
                    f"Setting up a paginated selector with {n} "
                    f"options spread across {page_count} pages."
                )
                self._setup_paginator(ctx, page_count)
            case n if n > cls.MAX_BUTTONS_PER_ROW:
                self.log(f"Setting up a simple dropdown menu with {n} options.")
                self._setup_simple_select(select_placeholder)
            case n if n >= cls.MIN_OPTIONS:
                self.log(f"Setting up an action row of {n} buttons.")
                self._setup_simple_buttons(button_style)
            case _:
                raise ValueError(f"At least {cls.MIN_OPTIONS} options are required.")

    def _setup_simple_buttons(self, style: ButtonStyle) -> None:
        for key in self.options:
            button = Button(style=style, label=key)
            button.callback = functools.partial(self.finish, button.label)
            self.add_item(button)

    def _setup_simple_select(self, placeholder: str | None) -> None:
        select = Select(
            placeholder=placeholder,
            options=[SelectOption(label=key) for key in self.options],
        )

        async def callback(interaction: Interaction) -> None:
            await self.finish(select.values[0], interaction)

        select.callback = callback
        self.add_item(select)

    def _setup_paginator(self, ctx: ApplicationContext, page_count: int) -> None:
        option_keys = list(self.options)
        quot, rmdr = divmod(len(self.options), page_count)

        def get_options(start: int, end: int) -> list[SelectOption]:
            return [SelectOption(label=key) for key in option_keys[start:end]]

        options_by_page = [
            get_options((i * quot) + min(i, rmdr), ((i + 1) * quot) + min(i + 1, rmdr))
            for i in range(page_count)
        ]
        Paginator(*options_by_page).attach(self, ctx)



from setuptools import setup

setup(
    url="https://github.com/nuztalgia/bot-ui-kitty",
    project_urls={
        "Issue Tracker": "https://github.com/nuztalgia/bot-ui-kitty/issues",
        "Source Code": "https://github.com/nuztalgia/bot-ui-kitty/tree/main/uikitty",
    },
    install_requires=[
        "py-cord ==2.2.2",
    ],
    use_scm_version={
        "local_scheme": lambda version: str(version.tag),
        "version_scheme": lambda _: "",
    },
)



from importlib.metadata import version
from pathlib import Path
from typing import Any, Final

from botstrap import Botstrap, CliColors, Color, Option
from discord import Bot


class UiKittyBot(Bot):
    def __init__(self, parent_dir: Path, force_sync: bool, **options: Any) -> None:
        super().__init__(**options)
        self._force_sync: Final[bool] = force_sync

        for file_path in parent_dir.glob("*.py"):
            if file_path.stem != "bot":
                print(f"Loading extension '{file_path.stem}'.")
                self.load_extension(file_path.stem)

    async def on_ready(self) -> None:
        if self._force_sync:
            print("Force-syncing commands. Be mindful of the rate limit.")
            await self.sync_commands(force=True)

        print("UI Kitty is online and ready!")


def main() -> int:
    parent_dir = Path(__file__).parent

    botstrap = Botstrap(
        name := "bot-ui-kitty",
        version=version(name),
        colors=CliColors(primary=Color.pink),
    ).register_token("main", storage_directory=parent_dir)

    args = botstrap.parse_args(
        force_sync=Option(flag=True, help="Force-sync application commands."),
    )
    botstrap.run_bot(
        bot_class=UiKittyBot,
        parent_dir=parent_dir,
        force_sync=args.force_sync,
    )
    return 0


if __name__ == "__main__":
    raise SystemExit(main())
